////////////////////////////////////////////////////////////////////////////////
// This file is part of the exemplary solution of Lab3 of the course
// "Software Engineering using Formal Methods" (2005)
// and playground for the verification tasks of Lab4
////////////////////////////////////////////////////////////////////////////////

/**
 * Class whose objects represent the automated teller machines of our bank.
 * These machines can operate both online and offline. In the first case, all
 * transactions are directly transfered to the central host of the bank, in the
 * latter case transactions are first buffered (using class
 * <code>OfflineAccountProxy</code> and later written back to the central
 * host.
 * 
 * @invariants accountProxies <> null
 *             and accountProxies.length = maxAccountNumber
 *             and centralHost <> null
 *             and centralHost.accounts <> null
 *             and centralHost.accounts.length = maxAccountNumber
 *             and ( customerAuthenticated implies insertedCard <> null )
 *             and ( insertedCard <> null implies not insertedCard.invalid )
 *             and ( insertedCard <> null
 *                   implies
 *                   insertedCard.accountNumber >= 0 and
 *                   insertedCard.accountNumber < maxAccountNumber )
 *             and ( online and insertedCard <> null implies
 *                   centralHost.accounts.get(insertedCard.accountNumber) <> null )

 */
public class ATM {
    
    static final int maxAccountNumber = CentralHost.maxAccountNumber;    

    /**
     * Currently instantiated offline account representatives. For simplicity,
     * here just an array is used to implement a map from account number the
     * <code>OfflineAccountProxy</code> objects. In practice this should
     * rather be a hashmap or something similar
     */
    private final OfflineAccountProxy[] accountProxies =
        new OfflineAccountProxy [maxAccountNumber];

    /**
     * The central host of the bank. We suppose that the attribute is only
     * accessed if the ATM is online to abstract from the actual communication 
     */
    private final CentralHost centralHost;
    
    /**
     * <code>true</code> iff the ATM is online, i.e. if
     * <code>centralHost</code> may be accessed
     */
    private boolean online = true;
    
    /**
     * A bank card that is possible inserted into the ATM at the time
     */
    private BankCard                    insertedCard          = null;
    /**
     * <code>true</code> iff there is a card inserted and the customer has
     * entered the right PIN
     */
    private boolean                     customerAuthenticated = false;
    /**
     * Count the number of wrong PINs entered since a valid card was inserted
     */
    private int                         wrongPINCounter       = 0;
    
    public ATM (CentralHost centralHost) {
        this.centralHost = centralHost;
    }

    
    /**
     * @return <code>true</code> iff the ATM is online
     * 
     * @stereotype query
     */
    private boolean isOnline () {
        return online;
    }
    
    
    /**
     * Access the account with number <code>accountNumber</code>. This either
     * returns the permanent account objects that is stored by the central host,
     * or an offline proxy
     * 
     * @param accountNumber
     *            the number of the account to be accessed
     * @return a representation of the account
     * 
     * TODO: The precondition can in practice not be ensured by the ATM (if it
     * is offline)
     *
     * TODO: the condition
     * "accountProxies.get(accountNumber).accountNumber =
     * accountNumber" this should really be an invariant and should
     * hold for all accounts, but how to specify this using KeY OCL???
     * 
     * @preconditions accountNumber >= 0 and accountNumber < maxAccountNumber
     *                and centralHost.accounts.get(accountNumber) <> null
     *                and centralHost.accounts.get(accountNumber).getAccountNumber()
                                        = accountNumber
     *                and ( accountProxies.get(accountNumber) <> null implies
     *                      accountProxies.get(accountNumber).getAccountNumber()
                                        = accountNumber )
     * @postconditions result.accountNumber = accountNumber
     */
    private Account getAccount (int accountNumber) {
        if ( isOnline () ) return centralHost.getAccount ( accountNumber );

        if ( !proxyExists ( accountNumber ) )
            setProxy ( accountNumber, new OfflineAccountProxy ( accountNumber ) );

        return getProxy ( accountNumber );
    }


    /**
     * Switch the status of <code>this</code> ATM to online or offline.
     * Switching to online will make the terminal write back all buffered
     * transactions to the central host and remove all account proxies.
     * 
     * TODO: one of the places where both complete specification and
     * verification are non-trivial
     * 
     * @param newOnline
     *            <code>true</code> iff the terminal is switched online
     * 
     * @postconditions online = newOnline
     */
    public void setOnline (boolean newOnline) {
        if ( this.online == newOnline ) return;
        
        if ( newOnline ) {
            System.out.println("ATM turned online ...");
            
            // synchronize with central host
            for ( int i = 0; i != CentralHost.maxAccountNumber; ++i ) {
                if ( proxyExists ( i ) ) {
                    getProxy ( i ).replay ( centralHost.getAccount ( i ) );
                    setProxy ( i, null );
                }
            }
        } else {
            System.out.println("ATM turned offline ...");            
        }
        
        if ( cardForNonexistingAccountInserted () ) confiscateCard ();
        
        this.online = newOnline;
    }
    
    
    ////////////////////////////////////////////////////////////////////////////
    // Methods for authentication of customers: Insertion of bank cards,
    // entering of PINs, etc.
    //
    // The public methods are written more defensively and throw a <code>
    // RuntimeException </code> in case of violated preconditions
    ////////////////////////////////////////////////////////////////////////////
    
    /**
     * Insert the given card into the ATM. This will check whether the card is
     * invalid or the card belongs to a non-existing account (if the terminal is
     * online); in this case the card is directly confiscated
     * 
     * @param card
     *            the card to be inserted
     *
     * @preconditions insertedCard = null
     *                and card <> null
     *                and card.accountNumber >= 0
     *                and card.accountNumber < maxAccountNumber
     * @postconditions ( card.invalid implies insertedCard = null )
     *                 and not customerAuthenticated
     *                 and wrongPINCounter <= 0
     */
    public void insertCard (BankCard card) {
        if ( !( !cardIsInserted () && card != null
                && card.getAccountNumber () >= 0
                && card.getAccountNumber () < maxAccountNumber ) )
                                throw new RuntimeException ();
        
        insertedCard = card;
        customerAuthenticated = false;
        wrongPINCounter = 0;
        if ( insertedCard.cardIsInvalid ()
             || cardForNonexistingAccountInserted () )
            confiscateCard ();
    }
    
    
    /**
     * Eject an inserted card
     * 
     * @return the card that was inserted previously
     *
     * @preconditions insertedCard <> null
     * @postconditions insertedCard = null
     */
    public BankCard ejectCard () {
        if ( !cardIsInserted() ) throw new RuntimeException ();

        final BankCard res = insertedCard;
        insertedCard = null;
        customerAuthenticated = false;
        
        return res;
    }
    
    
    /**
     * Confiscate an inserted card. This invalidates the card; afterwards, the
     * card is no longer regarded as inserted
     * 
     * @preconditions insertedCard <> null
     * @postconditions insertedCard = null and insertedCard@pre.invalid
     */
    public void confiscateCard () {
        if ( !cardIsInserted() ) throw new RuntimeException ();

        insertedCard.makeCardInvalid ();
        insertedCard = null;
        customerAuthenticated = false;        
    }
    
    
    /**
     * Enter the PIN the belongs to the currently inserted bank card into the
     * ATM. If a wrong PIN is entered three times in a row, the card is
     * confiscated. After having entered the correct PIN, the customer is
     * regarded is authenticated
     * 
     * @param pin
     *            the PIN that is entered
     *
     * TODO: postcondition is ugly
     * 
     * @preconditions insertedCard <> null and not customerAuthenticated
     * @postconditions if pin = insertedCard@pre.correctPIN then
     *                       customerAuthenticated
     *                 else
     *                       if wrongPINCounter@pre < 2 then
     *                              wrongPINCounter = wrongPINCounter@pre + 1
     *                       else
     *                              insertedCard = null
     *                              and insertedCard@pre.invalid
     *                       endif
     *                 endif
     */
    public void enterPIN (int pin) {
        if ( !( cardIsInserted () && !customerIsAuthenticated () ) )
                                    throw new RuntimeException ();

        if ( insertedCard.pinIsCorrect ( pin ) ) {
            customerAuthenticated = true;
        } else {
            ++wrongPINCounter;
            if ( wrongPINCounter >= 3 ) confiscateCard ();
        }
    }
    
    
    /**
     * @return <code>true</code> iff a card is inserted in the ATM
     * 
     * @stereotype query
     */
    private boolean cardIsInserted () {
        return insertedCard != null;
    }
    
    
    /**
     * @return <code>true</code> iff a customer is regarded as authenticated,
     *         i.e. if a valid card is inserted and the correct PIN has be
     *         entered
     *
     * @stereotype query
     */
    private boolean customerIsAuthenticated () {
        return customerAuthenticated;
    }

    
    /**
     * Check whether the currently inserted card belongs to an account that does
     * actually not exist
     * 
     * TODO: great postcondition ...
     * 
     * @return <code>true</code> iff there is a card inserted, the ATM is
     *         online and the card belongs to a non-existing account
     * 
     * @postconditions result = ( insertedCard <> null
     *                            and online
     *          and centralHost.accounts.get(insertedCard.accountNumber) = null )
     */
    private boolean cardForNonexistingAccountInserted () {
        return
            cardIsInserted ()
            && isOnline ()
            && !centralHost.accountExists ( getAccountNumber () );
    }
    
    
    
    ////////////////////////////////////////////////////////////////////////////
    // Methods for accessing an account after successful authentication
    //
    // The public methods are written more defensively and throw a <code>
    // RuntimeException </code> in case of violated preconditions
    ////////////////////////////////////////////////////////////////////////////
    
    /**
     * @return the account number of the account that belongs to the currently
     *         inserted card
     * 
     * @preconditions customerAuthenticated
     * @stereotype query
     */
    private int getAccountNumber () {
        return insertedCard.getAccountNumber ();
    }

    /**
     * @return the account that belongs to the currently inserted card
     * 
     * @preconditions customerAuthenticated
     */
    private Account getAccount () {
        return getAccount ( getAccountNumber () );
    }

    
    /**
     * Determine the balance of the current account.
     * 
     * @return the balance of the current account, provided that the balance is
     *         available (only in the online case). Otherwise <code>0</code>
     *         is returned
     * 
     * @preconditions customerAuthenticated
     * @postconditions online implies result =
     *                 centralHost.accounts.get(insertedCard.accountNumber).accountBalance()
     */
    public int accountBalance () {
        if ( !customerIsAuthenticated () ) throw new RuntimeException ();
        if ( !getAccount ().balanceIsAccessible () ) return 0;
        return getAccount ().accountBalance ();
    }
    
    
    /**
     * Try to withdraw <code>amount</code> from the current account.
     * 
     * TODO: make specification stronger, say something about the return value?
     * 
     * @param amount
     *            the amount that is supposed to be withdrawn
     * 
     * @return the amount of money that was sucessfully withdrawn (i.e. in case
     *         of failure the return value is <code>0</code>)
     * 
     * @preconditions customerAuthenticated and amount > 0
     */
    public int withdraw (int amount) {
        if ( !( customerIsAuthenticated () && amount > 0 ) )
                                    throw new RuntimeException ();
        return getAccount ().checkAndWithdraw ( amount );
    }

    
    /**
     * Request a printed account statement for the current account
     * 
     * @preconditions customerAuthenticated
     */
    public void requestAccountStatement () {
        if ( !customerIsAuthenticated () ) throw new RuntimeException ();
        getAccount ().requestStatement ();
    }

    
    
    ////////////////////////////////////////////////////////////////////////////
    // Internal methods for controlling account proxies
    ////////////////////////////////////////////////////////////////////////////
    
    /**
     * Update the mapping of account numbers to account proxies
     * 
     * @preconditions accountNumber >= 0 and accountNumber < maxAccountNumber
     * @postconditions accountProxies.get(accountNumber) = proxy
     */
    private void setProxy (int accountNumber, OfflineAccountProxy proxy) {
        accountProxies[accountNumber] = proxy;
    }

    /**
     * Map an account number to a possibly stored account proxy
     * 
     * @return the account proxy for account number <code>accountNumber</code>,
     *         or <code>null</code> if no such proxy exists
     * 
     * @stereotype query
     * @preconditions accountNumber >= 0 and accountNumber < maxAccountNumber
     */
    private OfflineAccountProxy getProxy (int accountNumber) {
        return accountProxies[accountNumber];
    }

    /**
     * @return <code>true</code> if there is a proxy for account number
     *         <code>accountNumber</code> stored in this object
     * 
     * @stereotype query
     * @preconditions accountNumber >= 0 and accountNumber < maxAccountNumber
     */
    private boolean proxyExists (int accountNumber) {
        return getProxy ( accountNumber ) != null;
    }
}
